#include <nsIFile.h>
#include <stdio.h>
#include <nsStringAPI.h>
#include <nsGenericFactory.h>
#include <nsIComponentRegistrar.h>
#include <nsIProperties.h>
#include <nsIComponentManager.h>
#include <nsDirectoryServiceUtils.h>
#include <prenv.h>
#include <plstr.h>

#define CELLML_PREPARER_CID                          \
{ /* 1dd113dc-268c-42b4-afa8-d6e03d4766ef */         \
    0x1dd113dc,                                      \
    0x268c,                                          \
    0x42b4,                                          \
    {0xaf, 0xa8, 0xd6, 0xe0, 0x3d, 0x47, 0x66, 0xef} \
}

class CellMLPreparer
  : public nsISupports /*,
                         public mozICellMLPreparer */
{
public:
  CellMLPreparer();
  NS_DECL_ISUPPORTS
protected:
  virtual ~CellMLPreparer() {}
public:
  PRBool mPrepareOk;
};

NS_IMPL_ISUPPORTS0(CellMLPreparer)


static PRBool
DoPreparations(nsIFile* pdir)
{
#ifdef XP_WIN
  char* envstr;
  nsAutoString path;
  nsCAutoString cpath;

  pdir->GetPath(path);
  CopyUTF16toUTF8(path, cpath);
  envstr = PR_GetEnv("PATH");
  // On Win32, libraries are picked up from the same directory as the
  // executable, and so Mozilla doesn't put the XPCOM/NSPR executables into the
  // path. We need this for the integrator/fitter to work...
  
  nsCOMPtr<nsIProperties>
    dsp(do_GetService(NS_DIRECTORY_SERVICE_CONTRACTID));
  if (dsp == (nsIProperties*)nsnull)
  {
    return PR_FALSE;
  }
  nsCOMPtr<nsIFile> cdir, cdir2;
  nsresult rv =
    dsp->Get("ComsD", NS_GET_IID(nsIFile), getter_AddRefs(cdir));
  if (cdir == nsnull)
  {
    printf("Can't get hold of the components directory.\n");
    return PR_FALSE;
  }
  cdir->GetParent(getter_AddRefs(cdir2));
  if (cdir2 == nsnull)
  {
    printf("Can't get hold of the components directory parent.\n");
    return PR_FALSE;
  }

  cdir2->GetPath(path);
  nsCAutoString cdpath;
  CopyUTF16toUTF8(path, cdpath);
  cpath += NS_LITERAL_CSTRING(";");
  cpath += cdpath;
  fflush(stdout);

  nsCAutoString cpt(NS_LITERAL_CSTRING("PATH="));
  if (envstr == nsnull)
  {
    cpt += cpath;
    PR_SetEnv(PL_strdup(cpt.get()));
  }
  else
  {
    cpt += cpath;
    cpt += NS_LITERAL_CSTRING(";");
    cpt += envstr;
    PR_SetEnv(PL_strdup(cpt.get()));
  }
#else
  // We need to copy the files to the Mozilla directory...
  nsCOMPtr<nsIProperties>
    dsp(do_GetService(NS_DIRECTORY_SERVICE_CONTRACTID));
  if (dsp == nsnull)
  {
    printf("Directory service could not be retrieved.\n");
    return PR_FALSE;
  }
  nsCOMPtr<nsIFile> cdir, cdir1;
  nsresult rv =
    dsp->Get("ComsD", NS_GET_IID(nsIFile), getter_AddRefs(cdir));
  if (rv != NS_OK || cdir == nsnull)
  {
    printf("Component directory could not be retrieved.\n");
    return PR_FALSE;
  }
  cdir->GetParent(getter_AddRefs(cdir1));
  if (cdir1 == nsnull)
  {
    printf("Component directory parent could not be retrieved.\n");
    return PR_FALSE;
  }

  // We want to copy the files into cdir...
  nsCOMPtr<nsIFile> fname;
  cdir1->Clone(getter_AddRefs(fname));
  fname->Append(NS_LITERAL_STRING("libomniORB4.so.0"));
  PRBool e;
  fname->Exists(&e);
  if (e)
  {
    return PR_TRUE;
  }

  pdir->Clone(getter_AddRefs(fname));
  fname->Append(NS_LITERAL_STRING("libomniORB4.so.0"));
  if (fname->CopyTo(cdir1, NS_LITERAL_STRING("")) != NS_OK)
  {
    return PR_FALSE;
  }

  pdir->Clone(getter_AddRefs(fname));
  fname->Append(NS_LITERAL_STRING("libomnithread.so.3"));
  if (fname->CopyTo(cdir1, NS_LITERAL_STRING("")) != NS_OK)
  {
    return PR_FALSE;
  }

  pdir->Clone(getter_AddRefs(fname));
  fname->Append(NS_LITERAL_STRING("libxpcorba.so"));
  if (fname->CopyTo(cdir1, NS_LITERAL_STRING("")) != NS_OK)
  {
    return PR_FALSE;
  }
#endif
  return PR_TRUE;
}

static NS_IMETHODIMP
CellMLPreparerRegistration
(
 nsIComponentManager *aCompMgr,
 nsIFile *aPath,
 const char *registryLocation,
 const char *componentType,
 const nsModuleComponentInfo *info
)
{
  nsCOMPtr<nsIFile> tdir, pdir;
  aPath->Clone(getter_AddRefs(pdir));
  pdir->GetParent(getter_AddRefs(tdir));
  tdir->GetParent(getter_AddRefs(pdir));
  DoPreparations(pdir);

  // It is now safe to load the remaining components...
  nsCOMPtr<nsIComponentRegistrar> ncr;
  aCompMgr->QueryInterface(NS_GET_IID(nsIComponentRegistrar),
                           getter_AddRefs(ncr));
  pdir->Append(NS_LITERAL_STRING("components_needcorba"));
  ncr->AutoRegister(pdir);
  return NS_OK;
}

static NS_IMETHODIMP
CellMLPreparerConstructor
(
 nsISupports* aOuter, REFNSIID aIID, void** aResult
)
{
  nsresult rv;
  *aResult = nsnull;

  if (aOuter != nsnull)
    return NS_ERROR_NO_AGGREGATION;

  CellMLPreparer *cp = new CellMLPreparer();
  if (cp == nsnull)
    return NS_ERROR_OUT_OF_MEMORY;
  if (!cp->mPrepareOk)
  {
    NS_ADDREF(cp);
    NS_RELEASE(cp);
    return NS_ERROR_FAILURE;
  }

  NS_ADDREF(cp);
  rv = cp->QueryInterface(aIID, aResult);
  NS_RELEASE(cp);

  return rv;
}

CellMLPreparer::CellMLPreparer()
  : mPrepareOk(PR_FALSE)
{
#ifdef XP_WIN
  setvbuf(stdout, nsnull, _IONBF, 0);
  setvbuf(stderr, nsnull, _IONBF, 0);
#endif

  nsCOMPtr<nsIProperties>
    dsp(do_GetService(NS_DIRECTORY_SERVICE_CONTRACTID));
  if (dsp == nsnull)
  {
    return;
  }
  nsCOMPtr<nsIFile> pdir;
  nsresult rv =
    dsp->Get("ProfD", NS_GET_IID(nsIFile), getter_AddRefs(pdir));
  /* Don't crash if it doesn't work... */
  if (rv != NS_OK || pdir == nsnull)
  {
    return;
  }
  pdir->Append(NS_LITERAL_STRING("extensions"));
  pdir->Append(NS_LITERAL_STRING("{ec337bea-b97f-4be2-8bc9-2b676146e43c}"));

  mPrepareOk = DoPreparations(pdir);
}

static const nsModuleComponentInfo components[] = 
{ 
  {
    "CellMLPreparer",
    CELLML_PREPARER_CID,
    "@cellml.org/cellml-preparer;1",
    CellMLPreparerConstructor,
    CellMLPreparerRegistration
  }
};

NS_IMPL_NSGETMODULE(CellMLModule,components)
